""" gather functions necessary to build an index """

import logging
from typing import Dict, Optional, Tuple, Union, Callable, Any, List

import faiss
import pandas as pd
from embedding_reader import EmbeddingReader

from autofaiss.external.metadata import IndexMetadata
from autofaiss.external.optimize import check_if_index_needs_training, get_optimal_index_keys_v2, get_optimal_train_size
from autofaiss.utils.cast import cast_bytes_to_memory_string, cast_memory_to_bytes, to_readable_time
from autofaiss.utils.decorators import Timeit
from autofaiss.indices import distributed
from autofaiss.indices.index_utils import initialize_direct_map
from autofaiss.indices.training import create_and_train_new_index
from autofaiss.indices.build import add_embeddings_to_index_local


logger = logging.getLogger("autofaiss")


def estimate_memory_required_for_index_creation(
    nb_vectors: int,
    vec_dim: int,
    index_key: Optional[str] = None,
    max_index_memory_usage: Optional[str] = None,
    make_direct_map: bool = False,
    nb_indices_to_keep: int = 1,
) -> Tuple[int, str]:
    """
    Estimates the RAM necessary to create the index
    The value returned is in Bytes
    """

    if index_key is None:
        if max_index_memory_usage is not None:
            index_key = get_optimal_index_keys_v2(
                nb_vectors, vec_dim, max_index_memory_usage, make_direct_map=make_direct_map
            )[0]
        else:
            raise ValueError("you should give max_index_memory_usage value if no index_key is given")

    metadata = IndexMetadata(index_key, nb_vectors, vec_dim, make_direct_map)

    index_memory = metadata.estimated_index_size_in_bytes()
    needed_for_adding = min(index_memory * 0.1, 10**9)

    index_needs_training = check_if_index_needs_training(index_key)

    if index_needs_training:
        # Compute the smallest number of vectors required to train the index given
        # the maximal memory constraint
        nb_vectors_train = get_optimal_train_size(nb_vectors, index_key, "1K", vec_dim)

        memory_for_training = metadata.compute_memory_necessary_for_training(nb_vectors_train)
    else:
        memory_for_training = 0

    # the calculation for max_index_memory_in_one_index comes from the way we split batches
    # see _batch_loader in distributed.py
    max_index_memory_in_one_index = index_memory // nb_indices_to_keep + index_memory % nb_indices_to_keep

    return int(max(max_index_memory_in_one_index + needed_for_adding, memory_for_training)), index_key


def get_estimated_construction_time_infos(nb_vectors: int, vec_dim: int, indent: int = 0) -> str:
    """
    Gives a general approximation of the construction time of the index
    """

    size = 4 * nb_vectors * vec_dim

    train = 1000  # seconds, depends on the number of points for training
    add = 450 * size / (150 * 1024**3)  # seconds, Linear approx (450s for 150GB in classic conditions)

    infos = (
        f"-> Train: {to_readable_time(train, rounding=True)}\n"
        f"-> Add: {to_readable_time(add, rounding=True)}\n"
        f"Total: {to_readable_time(train + add, rounding=True)}"
    )
    tab = "\t" * indent
    infos = tab + infos.replace("\n", "\n" + tab)

    return infos


def add_embeddings_to_index(
    embedding_reader: EmbeddingReader,
    trained_index_or_path: Union[str, faiss.Index],
    metadata: IndexMetadata,
    current_memory_available: str,
    embedding_ids_df_handler: Optional[Callable[[pd.DataFrame, int], Any]] = None,
    distributed_engine: Optional[str] = None,
    temporary_indices_folder: str = "hdfs://root/tmp/distributed_autofaiss_indices",
    nb_indices_to_keep: int = 1,
    index_optimizer: Optional[Callable] = None,
) -> Tuple[Optional[faiss.Index], Optional[Dict[str, str]]]:
    """Add embeddings to the index"""

    with Timeit("-> Adding the vectors to the index", indent=2):
        # Estimate memory available for adding embeddings to index
        size_per_index = metadata.estimated_index_size_in_bytes() / nb_indices_to_keep
        memory_available_for_adding = cast_bytes_to_memory_string(
            cast_memory_to_bytes(current_memory_available) - size_per_index
        )
        logger.info(
            f"The memory available for adding the vectors is {memory_available_for_adding}"
            "(total available - used by the index)"
        )

        if distributed_engine is None:
            return add_embeddings_to_index_local(
                embedding_reader=embedding_reader,
                trained_index_or_path=trained_index_or_path,
                memory_available_for_adding=memory_available_for_adding,
                embedding_ids_df_handler=embedding_ids_df_handler,
                index_optimizer=index_optimizer,
                add_embeddings_with_ids=False,
            )

        elif distributed_engine == "pyspark":
            return distributed.add_embeddings_to_index_distributed(
                trained_index_or_path=trained_index_or_path,
                embedding_reader=embedding_reader,
                memory_available_for_adding=memory_available_for_adding,
                embedding_ids_df_handler=embedding_ids_df_handler,
                temporary_indices_folder=temporary_indices_folder,
                nb_indices_to_keep=nb_indices_to_keep,
                index_optimizer=index_optimizer,
            )
        else:
            raise ValueError(f'Distributed by {distributed_engine} is not supported, only "pyspark" is supported')


def create_index(
    embedding_reader: EmbeddingReader,
    index_key: str,
    metric_type: Union[str, int],
    current_memory_available: str,
    embedding_ids_df_handler: Optional[Callable[[pd.DataFrame, int], Any]] = None,
    use_gpu: bool = False,
    make_direct_map: bool = False,
    distributed_engine: Optional[str] = None,
    temporary_indices_folder: str = "hdfs://root/tmp/distributed_autofaiss_indices",
    nb_indices_to_keep: int = 1,
    index_optimizer: Optional[Callable] = None,
) -> Tuple[Optional[faiss.Index], Optional[Dict[str, str]]]:
    """
    Create an index and add embeddings to the index
    """

    metadata = IndexMetadata(index_key, embedding_reader.count, embedding_reader.dimension, make_direct_map)

    # Create and train index
    trained_index = create_and_train_new_index(
        embedding_reader, index_key, metadata, metric_type, current_memory_available, use_gpu
    )

    # Add embeddings to index
    index, metrics = add_embeddings_to_index(
        embedding_reader,
        trained_index,
        metadata,
        current_memory_available,
        embedding_ids_df_handler,
        distributed_engine,
        temporary_indices_folder,
        nb_indices_to_keep,
        index_optimizer,
    )

    if make_direct_map:
        initialize_direct_map(index)

    return index, metrics


def create_partitioned_indexes(
    partitions: List[str],
    output_root_dir: str,
    embedding_column_name: str = "embedding",
    index_key: Optional[str] = None,
    index_path: Optional[str] = None,
    id_columns: Optional[List[str]] = None,
    should_be_memory_mappable: bool = False,
    max_index_query_time_ms: float = 10.0,
    max_index_memory_usage: str = "16G",
    min_nearest_neighbors_to_retrieve: int = 20,
    current_memory_available: str = "32G",
    use_gpu: bool = False,
    metric_type: str = "ip",
    nb_cores: Optional[int] = None,
    make_direct_map: bool = False,
    temp_root_dir: str = "hdfs://root/tmp/distributed_autofaiss_indices",
    big_index_threshold: int = 5_000_000,
    nb_splits_per_big_index: int = 1,
    maximum_nb_threads: int = 256,
) -> List[Optional[Dict[str, str]]]:
    """
    Create partitioned indexes from a list of parquet partitions, i.e. create one index per parquet partition

    Only supported with Pyspark. An active PySpark session must exist before calling this method
    """

    return distributed.create_partitioned_indexes(
        partitions=partitions,
        big_index_threshold=big_index_threshold,
        output_root_dir=output_root_dir,
        nb_cores=nb_cores,
        nb_splits_per_big_index=nb_splits_per_big_index,
        id_columns=id_columns,
        max_index_query_time_ms=max_index_query_time_ms,
        min_nearest_neighbors_to_retrieve=min_nearest_neighbors_to_retrieve,
        embedding_column_name=embedding_column_name,
        index_key=index_key,
        index_path=index_path,
        max_index_memory_usage=max_index_memory_usage,
        current_memory_available=current_memory_available,
        use_gpu=use_gpu,
        metric_type=metric_type,
        make_direct_map=make_direct_map,
        should_be_memory_mappable=should_be_memory_mappable,
        temp_root_dir=temp_root_dir,
        maximum_nb_threads=maximum_nb_threads,
    )



""" This file contain a class describing a memory efficient flat index """

import heapq
from typing import List, Optional, Tuple

from embedding_reader import EmbeddingReader

import faiss
import numpy as np
from tqdm import trange

from autofaiss.indices.faiss_index_wrapper import FaissIndexWrapper


class MemEfficientFlatIndex(FaissIndexWrapper):
    """
    Faiss-like Flat index that can support any size of vectors
    without memory issues.
    Two search functions are available to use either batch of smaller faiss
    flat index or rely fully on numpy.
    """

    def __init__(self, d: int, metric_type: int):
        """
        __init__ function for MemEfficientFlatIndex

        Parameters:
        -----------
        d : int
            dimension of the vectors, named d to keep Faiss notation
        metric_type : int
            similarity metric used in the vector space, using faiss
            enumerate values (faiss.METRIC_INNER_PRODUCT and faiss.METRIC_L2)
        """

        super().__init__(d, metric_type)

        self.dim = d
        self.prod_emb = np.zeros((0, self.dim))
        self.embedding_reader: Optional[EmbeddingReader] = None

    def delete_vectors(self):
        """delete the vectors of the index"""
        self.prod_emb = np.zeros((0, self.dim))

    # pylint: disable=missing-function-docstring, invalid-name
    def add(self, x: np.ndarray):
        if self.prod_emb.shape[0] == 0:
            self.prod_emb = x.astype(np.float32)
        else:
            raise NotImplementedError("You can add vectors only once, delete them first with delete_vectors")

    def add_all(self, filename: str, nb_items: int):
        """
        Function that adds vectors to the index from a memmory-mapped array

        Parameters
        ----------
        filename : string
            path of the 2D numpy array of shape (nb_items, vector_dim)
            on the disk
        nb_items : int
            number of vectors in the 2D array (the dim is already known)
        """

        if self.prod_emb.shape[0] == 0:
            self.prod_emb = np.memmap(filename, dtype="float32", mode="r", shape=(nb_items, self.dim))
        else:
            raise NotImplementedError("You can add vectors only once, delete them first")

    def add_files(self, embedding_reader: EmbeddingReader):
        if self.embedding_reader is None:
            self.embedding_reader = embedding_reader
        else:
            raise NotImplementedError("You can add vectors only once, delete them first with delete_vectors")

    # pylint: disable too_many_locals
    def search_numpy(self, xq: np.ndarray, k: int, batch_size: int = 4_000_000):
        """
        Function that search the k nearest neighbours of a batch of vectors.
        This implementation is based on vectorized numpy function, it is slower than
        the search function based on batches of faiss flat indices.
        We keep this implementation because we can build new functions using this code.
        Moreover, the distance computation is more precise in numpy than the faiss implementation
        that optimizes speed over precision.

        Parameters
        ----------
        xq : 2D numpy.array of floats
            Batch of vectors of shape (batch_size, vector_dim)
        k : int
            Number of neighbours to retrieve for every vector
        batch_size : int
            Size of the batch of vectors that are explored.
            A bigger value is prefered to avoid multiple loadings
            of the vectors from the disk.

        Returns
        -------
        D : 2D numpy.array of floats
            Distances numpy array of shape (batch_size, k).
            Contains the distances computed by the index of the k nearest neighbours.
        I : 2D numpy.array of ints
            Labels numpy array of shape (batch_size, k).
            Contains the vectors' labels of the k nearest neighbours.
        """
        assert self.metric_type == faiss.METRIC_INNER_PRODUCT

        # Instanciate several heaps, (is there a way to have vectorized heaps?)
        h: List[List[Tuple[float, int]]] = [[] for _ in range(xq.shape[0])]

        # reshape input for vectorized distance computation
        xq_reshaped = np.expand_dims(xq, 1)

        # initialize index offset
        offset = 0

        # For each batch
        for i in trange(0, self.prod_emb.shape[0], batch_size):
            # compute distances in one tensor product
            dist_arr = np.sum((xq_reshaped * np.expand_dims(self.prod_emb[i : i + batch_size], 0)), axis=-1)

            # get index of the k biggest
            # pylint: disable=unsubscriptable-object # pylint/issues/3139
            max_k = min(k, dist_arr.shape[1])

            ind_k_max = np.argpartition(dist_arr, -max_k)[:, -max_k:]

            assert ind_k_max.shape == (xq.shape[0], max_k)

            # to be vectorized if it is indeed the bottleneck, (it's not for batch_size >> 10000)
            for j, inds in enumerate(ind_k_max):
                for ind, distance in zip(inds, dist_arr[j, inds]):
                    true_ind = offset + ind if ind != -1 else -1
                    if len(h[j]) < k:
                        heapq.heappush(h[j], (distance, true_ind))
                    else:
                        heapq.heappushpop(h[j], (distance, true_ind))

            offset += batch_size

        # Fill distance and indice matrix
        D: np.ndarray = np.zeros((xq.shape[0], k), dtype=np.float32)
        I: np.ndarray = np.full((xq.shape[0], k), fill_value=-1, dtype=np.int32)

        for i in range(xq.shape[0]):
            # case where we couldn't find enough vectors
            max_k = min(k, len(h[i]))
            for j in range(max_k):
                x = heapq.heappop(h[i])
                D[i][max_k - 1 - j] = x[0]
                I[i][max_k - 1 - j] = x[1]

        return D, I

    # pylint: disable=too-many-locals, arguments-differ
    def search(self, x: np.ndarray, k: int, batch_size: int = 4_000_000):
        """
        Function that search the k nearest neighbours of a batch of vectors

        Parameters
        ----------
        x : 2D numpy.array of floats
            Batch of vectors of shape (batch_size, vector_dim)
        k : int
            Number of neighbours to retrieve for every vector
        batch_size : int
            Size of the batch of vectors that are explored.
            A bigger value is prefered to avoid multiple loadings
            of the vectors from the disk.

        Returns
        -------
        D : 2D numpy.array of floats
            Distances numpy array of shape (batch_size, k).
            Contains the distances computed by the index of the k nearest neighbours.
        I : 2D numpy.array of ints
            Labels numpy array of shape (batch_size, k).
            Contains the vectors' labels of the k nearest neighbours.
        """

        if self.prod_emb is None:
            raise ValueError("The index is empty")

        # Cast in the right format for Faiss
        if x.dtype != np.float32:
            x = x.astype(np.float32)

        # xq for x query, a better name than x which is Faiss convention
        xq = x

        # Instanciate several heaps, (is there a way to have vectorized heaps?)
        h: List[List[Tuple[float, int]]] = [[] for _ in range(xq.shape[0])]

        # initialize index offset
        offset = 0

        # For each batch
        for i in trange(0, self.prod_emb.shape[0], batch_size):
            # instanciate a Flat index
            brute = faiss.IndexFlatIP(self.dim)
            # pylint: disable=no-value-for-parameter
            brute.add(self.prod_emb[i : i + batch_size])
            D_tmp, I_tmp = brute.search(xq, k)

            # to be vectorized if it is indeed the bottleneck, (it's not for batch_size >> 10000)
            for j, (distances, inds) in enumerate(zip(D_tmp, I_tmp)):
                for distance, ind in zip(distances, inds):
                    true_ind: int = offset + ind if ind != -1 else -1
                    if len(h[j]) < k:
                        heapq.heappush(h[j], (distance, true_ind))
                    else:
                        heapq.heappushpop(h[j], (distance, true_ind))

            offset += batch_size

        # Fill distance and indice matrix
        D: np.ndarray = np.zeros((xq.shape[0], k), dtype=np.float32)
        I: np.ndarray = np.full((xq.shape[0], k), fill_value=-1, dtype=np.int32)

        for i in range(xq.shape[0]):
            # case where we couldn't find enough vectors
            max_k = min(k, len(h[i]))
            for j in range(max_k):
                x = heapq.heappop(h[i])
                D[i][max_k - 1 - j] = x[0]
                I[i][max_k - 1 - j] = x[1]

        return D, I

    def search_files(self, x: np.ndarray, k: int, batch_size: int):
        if self.embedding_reader is None:
            raise ValueError("The index is empty")

        # Cast in the right format for Faiss
        if x.dtype != np.float32:
            x = x.astype(np.float32)

        # xq for x query, a better name than x which is Faiss convention
        xq = x

        # Instanciate several heaps, (is there a way to have vectorized heaps?)
        h: List[List[Tuple[float, int]]] = [[] for _ in range(xq.shape[0])]

        # initialize index offset
        offset = 0

        # For each batch
        for emb_array, _ in self.embedding_reader(batch_size):
            # for i in trange(0, self.prod_emb.shape[0], batch_size):

            # instanciate a Flat index
            brute = faiss.IndexFlatIP(self.dim)
            # pylint: disable=no-value-for-parameter
            brute.add(emb_array)
            D_tmp, I_tmp = brute.search(xq, k)

            # to be vectorized if it is indeed the bottleneck, (it's not for batch_size >> 10000)
            for j, (distances, inds) in enumerate(zip(D_tmp, I_tmp)):
                for distance, ind in zip(distances, inds):
                    true_ind: int = offset + ind if ind != -1 else -1
                    if len(h[j]) < k:
                        heapq.heappush(h[j], (distance, true_ind))
                    else:
                        heapq.heappushpop(h[j], (distance, true_ind))

            offset += emb_array.shape[0]

        # Fill distance and indice matrix
        D: np.ndarray = np.zeros((xq.shape[0], k), dtype=np.float32)
        I: np.ndarray = np.full((xq.shape[0], k), fill_value=-1, dtype=np.int32)

        for i in range(xq.shape[0]):
            # case where we couldn't find enough vectors
            max_k = min(k, len(h[i]))
            for j in range(max_k):
                x = heapq.heappop(h[i])  # type: ignore
                D[i][max_k - 1 - j] = x[0]
                I[i][max_k - 1 - j] = x[1]

        return D, I



""" Functions to find optimal index parameters """
import json
import logging
import re
from functools import partial, reduce
from math import floor, log2, sqrt
from operator import mul
from typing import Callable, List, Optional, TypeVar

import faiss
import fsspec
import numpy as np
from autofaiss.external.metadata import IndexMetadata, compute_memory_necessary_for_training_wrapper
from autofaiss.external.scores import compute_fast_metrics
from autofaiss.indices.index_utils import set_search_hyperparameters, speed_test_ms_per_query
from autofaiss.utils.algorithms import discrete_binary_search
from autofaiss.utils.cast import cast_memory_to_bytes
from autofaiss.utils.decorators import Timeit
from embedding_reader import EmbeddingReader

logger = logging.getLogger("autofaiss")


def check_if_index_needs_training(index_key: str) -> bool:
    """
    Function that checks if the index needs to be trained
    """

    if "IVF" in index_key:
        return True
    elif "IMI" in index_key:
        return True
    else:
        return False


def index_key_to_nb_cluster(index_key: str) -> int:
    """
    Function that takes an index key and returns the number of clusters
    """

    matching = re.findall(r"IVF\d+|IMI\d+x\d+", index_key)

    if matching:
        # case IVF index
        if re.findall(r"IVF\d+", matching[0]):
            nb_clusters = int(matching[0][3:])
        # case IMI index
        elif re.findall(r"IMI\d+x\d+", matching[0]):
            nb_clusters = 2 ** reduce(mul, [int(num) for num in re.findall(r"\d+", matching[0])])
        else:
            raise ValueError(f"Unable to determine the number of clusters for index {index_key}")
    else:
        raise ValueError(f"Unable to determine the number of clusters for index {index_key}")

    return nb_clusters


def get_optimal_train_size(
    nb_vectors: int, index_key: str, current_memory_available: Optional[str], vec_dim: Optional[int]
) -> int:
    """
    Function that determines the number of training points necessary to
    train the index, based on faiss heuristics for k-means clustering.
    """

    matching = re.findall(r"IVF\d+|IMI\d+x\d+", index_key)

    if matching:
        nb_clusters = index_key_to_nb_cluster(index_key)
        points_per_cluster: int = 100

        # compute best possible number of vectors to give to train the index
        # given memory constraints
        if current_memory_available and vec_dim:
            memory_per_cluster_set = compute_memory_necessary_for_training_wrapper(
                points_per_cluster, index_key, vec_dim
            )
            size = cast_memory_to_bytes(current_memory_available)
            points_per_cluster = max(min(size / memory_per_cluster_set, points_per_cluster), 31.0)

        # You will need between 30 * nb_clusters and 256 * nb_clusters to train the index
        train_size = min(round(points_per_cluster * nb_clusters), nb_vectors)

    else:
        raise ValueError(f"Unknown index type: {index_key}")

    return train_size


def get_optimal_batch_size(vec_dim: int, current_memory_available: str) -> int:
    """compute optimal batch size to use the RAM at its full potential for adding vectors"""

    memory = cast_memory_to_bytes(current_memory_available)

    batch_size = int(min(memory, 10**9) / (vec_dim * 4))  # using more than 1GB of ram is not faster here

    return batch_size


def get_optimal_nb_clusters(nb_vectors: int) -> List[int]:
    """
    Returns a list with the recommended number of clusters for an index containing nb_vectors vectors.
    The first value is the most recommended one.
    see: https://github.com/facebookresearch/faiss/wiki/Guidelines-to-choose-an-index
    """

    nb_clusters_list = []

    if nb_vectors < 1_000_000:
        # no need to use HNSW for small number of clusters
        x_initial = 4 * sqrt(nb_vectors)  # between 4xsqrt(n) and 16xsqrt(n)
        x_closest_power = 2 ** round(log2(x_initial))
        nb_clusters_list.append(round(x_closest_power))
        nb_clusters_list.append(2 * x_closest_power)
        nb_clusters_list.append(x_closest_power)
        nb_clusters_list.append(round(x_initial))
    elif nb_vectors < 10_000_000:
        nb_clusters_list.append(16_384)
        nb_clusters_list.append(65_536)
    elif nb_vectors < 300_000_000:
        nb_clusters_list.append(65_536)
        nb_clusters_list.append(2**17)
        nb_clusters_list.append(2**18)  # slow training !
    else:
        nb_clusters_list.append(2**17)
        nb_clusters_list.append(2**18)  # slow training !
        nb_clusters_list.append(65_536)
        nb_clusters_list.append(2**20)  # very slow training !

    nb_clusters_list = [int(x) for x in nb_clusters_list]

    if not nb_clusters_list:
        return [256]  # default value

    return nb_clusters_list


def get_optimal_index_keys_v2(
    nb_vectors: int,
    dim_vector: int,
    max_index_memory_usage: str,
    flat_threshold: int = 1000,
    quantization_threshold: int = 10000,
    force_pq: Optional[int] = None,
    make_direct_map: bool = False,
    should_be_memory_mappable: bool = False,
    ivf_flat_threshold: int = 1_000_000,
    use_gpu: bool = False,
) -> List[str]:
    """
    Gives a list of interesting indices to try, *the one at the top is the most promising*

    See: https://github.com/facebookresearch/faiss/wiki/Guidelines-to-choose-an-index for
    detailed explanations.
    """

    # Exception cases:

    if nb_vectors < flat_threshold:  # When less than 1000 vectors, the flat index is usually faster
        return ["IVF1,Flat" if should_be_memory_mappable else "Flat"]
    if nb_vectors < quantization_threshold:  # quantization is not possible (>=10_000 vectors needed)
        if should_be_memory_mappable:
            return get_optimal_ivf(nb_vectors)
        return ["HNSW15"]
    if force_pq is not None:  # Forced quantization
        return get_optimal_quantization(nb_vectors, dim_vector, force_quantization_value=force_pq)

    # General cases:

    # Get max memory usage
    max_size_in_bytes = cast_memory_to_bytes(max_index_memory_usage)

    if not should_be_memory_mappable:
        # If we can build an HNSW with the given memory constraints, it's the best
        m_hnsw = int(floor((max_size_in_bytes / (4 * nb_vectors) - dim_vector) / 2))
        if m_hnsw >= 8:
            return [f"HNSW{min(m_hnsw, 32)}"]
    if nb_vectors < ivf_flat_threshold or use_gpu:
        # Try to build a not quantized IVF index
        index_keys = get_optimal_ivf(nb_vectors)
        index_metadata = IndexMetadata(index_keys[0], nb_vectors, dim_vector, make_direct_map)
        if index_metadata.estimated_index_size_in_bytes() <= max_size_in_bytes:
            return index_keys

    # Otherwise, there is not enough space, let's go for quantization
    return get_optimal_quantization(nb_vectors, dim_vector, force_max_index_memory_usage=max_index_memory_usage)


def get_optimal_ivf(nb_vectors: int) -> List[str]:
    """
    Function that returns a list of relevant index_keys to create not quantized IVF indices.

    Parameters
    ----------
    nb_vectors: int
        Number of vectors in the dataset.
    """
    index_keys = []

    for nb_clusters in get_optimal_nb_clusters(nb_vectors):
        index_keys.append(f"IVF{nb_clusters},Flat")

    return index_keys


def get_optimal_quantization(
    nb_vectors: int,
    dim_vector: int,
    force_quantization_value: Optional[int] = None,
    force_max_index_memory_usage: Optional[str] = None,
) -> List[str]:
    """
    Function that returns a list of relevant index_keys to create quantized indices.

    Parameters:
    ----------
    nb_vectors: int
        Number of vectors in the dataset.
    dim_vector: int
        Dimension of the vectors in the dataset.
    force_quantization_value: Optional[int]
        Force to use this value as the size of the quantized vectors (PQx).
        It can be used with the force_max_index_memory_usage parameter,
        but the result might be empty.
    force_max_index_memory_usage: Optional[str]
        Add a memory constraint on the index.
        It can be used with the force_quantization_value parameter,
        but the result might be empty.

    Return:
    -------
    index_keys: List[str]
        List of index_keys that would be good choices for quantization.
        The list can be empty if the given constraints are too strong.
    """

    # Default values
    pq_values = [256, 128, 64, 48, 32, 24, 16, 8, 4]
    targeted_compression_ratio = 0.0  # 0 = no constraint

    # Force compression ratio if required
    if force_max_index_memory_usage is not None:
        total_bytes = 4.0 * nb_vectors * dim_vector  # x4 because float32
        max_mem_bytes = float(cast_memory_to_bytes(force_max_index_memory_usage))
        targeted_compression_ratio = total_bytes / max_mem_bytes

    # Force quantization value if required
    if force_quantization_value is not None:
        pq_values = [force_quantization_value]

    # Compute optimal number of clusters
    relevant_list: List[str] = []
    nb_clusters_list = get_optimal_nb_clusters(nb_vectors)

    # Look for matching index keys
    for pq in pq_values:
        if pq < dim_vector:
            for nb_clusters in nb_clusters_list:
                # Compute quantized vector size

                # https://github.com/facebookresearch/faiss/blob/main/faiss/invlists/InvertedLists.h#L193
                embedding_id_byte = 8

                vector_size_byte = pq + embedding_id_byte

                # Compute compression ratio with quantization PQx
                compression_ratio = (4 * dim_vector) / vector_size_byte

                # Add index_key if compression ratio is high enough
                if compression_ratio >= targeted_compression_ratio:
                    # y is a multiple of pq (required)
                    # y <= d, with d the dimension of the input vectors (preferable)
                    # y <= 6*pq (preferable)
                    # here we choose a y slightly bigger than d to avoid losing information
                    # in case such as 101, 128 is better than 64 to avoid losing information
                    # in the linear transform
                    y = (min(dim_vector // pq, 6) + 1) * pq
                    cluster_opt = f"IVF{nb_clusters}" if nb_clusters < 1000 else f"IVF{nb_clusters}_HNSW32"
                    relevant_list.append(f"OPQ{pq}_{y},{cluster_opt},PQ{pq}x8")

    return relevant_list


T = TypeVar("T", int, float)


def get_min_param_value_for_best_neighbors_coverage(
    index: faiss.Index,
    parameter_range: List[T],
    hyperparameter_str_from_param: Callable[[T], str],
    targeted_nb_neighbors_to_query: int,
    *,
    targeted_coverage: float = 0.99,
    use_gpu: bool = False,
) -> T:
    """
    This function returns the minimal value to set in the index hyperparameters so that,
    on average, the index retrieves 99% of the requested k=targeted_nb_neighbors_to_query nearest neighbors.

            1 ^       ------------------------
              |     /
    nearest   |    /
    neighbors |   /
    coverage  |  /
              | /
            0 +--[--------------------------]-->  param_value
                 ^  ^                       ^
                 |  |                       |
                 |  min_param_value         |
                 |                          |
                 min(parameter_range)      max(parameter_range)

    Parameters
    ----------
    index: faiss.Index
        Index to search on.
    parameter_range: List[T]
        List of possible values for the hyperparameter. This list is sorted.
    hyperparameter_str_from_param: Callable[[T], str]
        Function to generate a hyperparameter string from the hyperparameter value
        on which we do a binary search.
    targeted_nb_neighbors_to_query: int
        Targeted number of neighbors to query.
    targeted_coverage: float
        Targeted nearest neighbors coverage. The average ratio of neighbors really retrived
        when asking for k=targeted_nb_neighbors_to_query nearest neighbors.
    use_gpu: bool
        Whether the index is on the GPU.
    """

    # Initialize query vectors to run the benchmark
    query_vectors = index.reconstruct_n(0, min(index.ntotal, 100))

    # Function to compute the coverage of the nearest neighbors
    def get_nearest_neighbors_coverage(k: int) -> float:
        ind = index.search(query_vectors, k)[1]
        return 1 - np.sum(ind == -1) / ind.size

    # Display a warning if the targeted number of nearest neighbors is incoherent with the index size
    if targeted_nb_neighbors_to_query > index.ntotal:
        logger.warning(
            f"The targeted number of nearest neighbors ({targeted_nb_neighbors_to_query}) "
            f"is greater than the total number of vectors in the index ({index.ntotal}). "
            "We set the targeted number of nearest neighbors to the total number of vectors."
        )
        targeted_nb_neighbors_to_query = index.ntotal

    # Compute the max nearest neighbors coverage possible with the given hyperparameters
    param_str = hyperparameter_str_from_param(parameter_range[-1])
    set_search_hyperparameters(index, param_str, use_gpu)
    max_nearest_neighbors_coverage = get_nearest_neighbors_coverage(targeted_nb_neighbors_to_query)

    # If the index cannot reach the targeted coverage, we adapt it.
    if max_nearest_neighbors_coverage < targeted_coverage:
        logger.warning(
            f"The maximum nearest neighbors coverage is {100*max_nearest_neighbors_coverage:.2f}% for this index. "
            f"It means that when requesting {targeted_nb_neighbors_to_query} nearest neighbors, the average number "
            f"of retrieved neighbors will be {round(targeted_nb_neighbors_to_query*max_nearest_neighbors_coverage)}. "
            f"The program will try to find the best hyperparameters to reach 95% of this max coverage at least, "
            "and then will optimize the search time for this target. "
            "The index search speed could be higher than the requested max search speed."
        )

        # In that case there is a hard limit on the maximal nearest neighbors coverage.
        # We redefine the new targeted coverage to reach the begining of the inflexion point
        #
        #          1 ^                               <---- Initial target: 99% coverage
        #            |
        #  nearest   |     ------------------------- <---- New target 0.95*max_nearest_neighbors_coverage
        #  neighbors |   /
        #  coverage  |  /
        #            | /
        #          0 +--[--------------------------]-->  param_value
        #               ^  ^                       ^
        #               |  |                       |
        #               |  min_param_value         |
        #               |                          |
        #               min(parameter_range)      max(parameter_range)

        targeted_coverage = 0.95 * max_nearest_neighbors_coverage

    # Intialize the binary search
    def is_meeting_constraint(rank: int) -> bool:
        parameter_value = parameter_range[rank]
        param_str = hyperparameter_str_from_param(parameter_value)
        set_search_hyperparameters(index, param_str, use_gpu)
        nearest_neighbors_coverage = get_nearest_neighbors_coverage(targeted_nb_neighbors_to_query)

        return nearest_neighbors_coverage >= targeted_coverage

    # Find the min param_value that reaches the targeted coverage
    best_rank = max(0, discrete_binary_search(is_meeting_constraint, len(parameter_range)) - 1)

    return parameter_range[best_rank]


def binary_search_on_param(
    index: faiss.Index,
    parameter_range: List[T],
    max_speed_ms: float,  # milliseconds
    hyperparameter_str_from_param: Callable[[T], str],
    timeout_boost_for_precision_search: float = 6.0,
    use_gpu: bool = False,
    max_timeout_per_iteration_s: float = 1.0,  # seconds
) -> T:
    """
    Apply a binary search on a given hyperparameter to maximize the recall given
    a query speed constraint in milliseconds/query.

    Parameters
    ----------
    index: faiss.Index
        Index to search on.
    parameter_range: List[T]
        List of possible values for the hyperparameter. This list is sorted.
    max_speed_ms: float
        Maximum query speed in milliseconds/query.
    hyperparameter_str_from_param: Callable[[T], str]
        Function to generate a hyperparameter string from the hyperparameter value
        on which we do a binary search.
    timeout_boost_for_precision_search: float
        Timeout boost for the precision search phase.
    use_gpu: bool
        Whether the index is on the GPU.
    max_timeout_per_iteration_s: float
        Maximum timeout per iteration in seconds.
    """

    query_vectors = index.reconstruct_n(0, min(index.ntotal, 4000))
    timout_s = 15 * max_speed_ms / 1000

    get_speed = partial(
        speed_test_ms_per_query, query=query_vectors, ksearch=40, timout_s=min(max_timeout_per_iteration_s, timout_s)
    )

    def is_not_acceptable_speed(rank: int) -> bool:
        parameter_value = parameter_range[rank]
        param_str = hyperparameter_str_from_param(parameter_value)
        set_search_hyperparameters(index, param_str, use_gpu)
        speed = get_speed(index)

        return speed >= max_speed_ms

    best_rank = max(0, discrete_binary_search(is_not_acceptable_speed, len(parameter_range)) - 1)

    # make sure that the query time is respected by spending X more time to evaluate the query speed
    decreasing_ratio = 0.95

    query_vectors = index.reconstruct_n(0, min(index.ntotal, 50000))
    get_speed = partial(
        speed_test_ms_per_query,
        query=query_vectors,
        ksearch=40,
        timout_s=min(max_timeout_per_iteration_s, timeout_boost_for_precision_search * timout_s),
    )

    while is_not_acceptable_speed(best_rank) and best_rank > 1:
        best_rank -= max(1, floor((1 - decreasing_ratio) * best_rank))

    best_rank = max(0, min(best_rank, len(parameter_range) - 1))

    return parameter_range[best_rank]


def get_optimal_hyperparameters(
    index,
    index_key: str,
    max_speed_ms: float,  # milliseconds
    use_gpu: bool = False,
    max_timeout_per_iteration_s: float = 1.0,  # seconds
    min_ef_search: int = 32,
    min_nearest_neighbors_to_retrieve: int = 20,
) -> str:
    """Find the optimal hyperparameters to maximize the recall given a query speed in milliseconds/query"""

    params = [int(x) for x in re.findall(r"\d+", index_key)]

    if any(re.findall(r"OPQ\d+_\d+,IVF\d+,PQ\d+", index_key)):
        ht = 2048
        nb_clusters = int(params[2])
        hyperparameter_str_from_param = (
            lambda nprobe: f"nprobe={nprobe},ht={ht}"  # pylint: disable=unnecessary-lambda-assignment
        )
        parameter_range = list(range(1, min(6144, nb_clusters) + 1))
        timeout_boost_for_precision_search = 6.0

    elif any(re.findall(r"OPQ\d+_\d+,IVF\d+_HNSW\d+,PQ\d+", index_key)):
        ht = 2048
        nb_clusters = int(params[2])
        hyperparameter_str_from_param = (
            lambda nprobe: f"nprobe={nprobe},efSearch={2*nprobe},ht={ht}"  # pylint: disable=unnecessary-lambda-assignment
        )
        parameter_range = list(range(max(1, min_ef_search // 2), min(6144, nb_clusters) + 1))
        timeout_boost_for_precision_search = 12.0

    elif any(re.findall(r"HNSW\d+", index_key)):
        hyperparameter_str_from_param = (
            lambda ef_search: f"efSearch={ef_search}"  # pylint: disable=unnecessary-lambda-assignment
        )
        parameter_range = list(range(16, 2**14))
        timeout_boost_for_precision_search = 6.0

    elif any(re.findall(r"IVF\d+,Flat", index_key)):
        nb_clusters = int(params[0])
        hyperparameter_str_from_param = (
            lambda nprobe: f"nprobe={nprobe}"  # pylint: disable=unnecessary-lambda-assignment
        )
        parameter_range = list(range(1, nb_clusters + 1))
        timeout_boost_for_precision_search = 6.0

    elif index_key == "Flat":
        return ""

    else:
        raise NotImplementedError(f"we don't have heuristics for that kind or index ({index_key})")

    min_param_value = get_min_param_value_for_best_neighbors_coverage(
        index, parameter_range, hyperparameter_str_from_param, min_nearest_neighbors_to_retrieve, use_gpu=use_gpu
    )

    parameter_range = [param_value for param_value in parameter_range if param_value >= min_param_value]
    parameter_range = parameter_range or [min_param_value]

    optimal_param = binary_search_on_param(
        index,
        parameter_range,
        max_speed_ms,
        hyperparameter_str_from_param,
        timeout_boost_for_precision_search,
        use_gpu,
        max_timeout_per_iteration_s,
    )

    return hyperparameter_str_from_param(optimal_param)


def optimize_and_measure_index(
    embedding_reader: EmbeddingReader,
    index: faiss.Index,
    index_infos_path: Optional[str],
    index_key: str,
    index_param: Optional[str],
    index_path: Optional[str],
    *,
    max_index_query_time_ms: float,
    min_nearest_neighbors_to_retrieve: int,
    save_on_disk: bool,
    use_gpu: bool,
):
    """Optimize one index by selecting the best hyperparameters and calculate its metrics"""
    if index_param is None:
        with Timeit(f"Computing best hyperparameters for index {index_path}", indent=1):
            index_param = get_optimal_hyperparameters(
                index,
                index_key,
                max_speed_ms=max_index_query_time_ms,
                min_nearest_neighbors_to_retrieve=min_nearest_neighbors_to_retrieve,
                use_gpu=use_gpu,
            )
    # Set search hyperparameters for the index
    set_search_hyperparameters(index, index_param, use_gpu)
    logger.info(f"The best hyperparameters are: {index_param}")
    metric_infos = {"index_key": index_key, "index_param": index_param, "index_path": index_path}
    with Timeit("Compute fast metrics", indent=1):
        metric_infos.update(compute_fast_metrics(embedding_reader, index))
    if save_on_disk:
        with Timeit("Saving the index on local disk", indent=1):
            with fsspec.open(index_path, "wb").open() as f:
                faiss.write_index(index, faiss.PyCallbackIOWriter(f.write))
            with fsspec.open(index_infos_path, "w").open() as f:
                json.dump(metric_infos, f)

    return metric_infos



import numpy as np
from autofaiss import build_index


def test_np_quantize():
    embs = np.ones((100, 512), "float32")
    index, _ = build_index(embs, save_on_disk=False)
    _, I = index.search(embs, 1)
    assert I[0][0] == 0



""" Various optimization algorithms """

from typing import Callable


# pylint: disable=invalid-name
def discrete_binary_search(is_ok: Callable[[int], bool], n: int) -> int:
    """
    Binary search in a function domain

    Parameters
    ----------
    is_ok : bool
        Boolean monotone function defined on range(n)
    n : int
        length of the search scope

    Returns
    -------
    i : int
        first index i such that is_ok(i) or returns n if is_ok is all False

    :complexity: O(log(n))
    """

    lo = 0
    hi = n
    while lo < hi:
        mid = lo + (hi - lo) // 2
        if mid >= n or is_ok(mid):
            hi = mid
        else:
            lo = mid + 1
    return lo



""" Functions to compute metrics on an index """
import fsspec
from typing import Dict, Union, Optional

import numpy as np
import faiss

from embedding_reader import EmbeddingReader

from autofaiss.indices.index_utils import get_index_size, search_speed_test
from autofaiss.indices.memory_efficient_flat_index import MemEfficientFlatIndex
from autofaiss.metrics.recalls import one_recall_at_r, r_recall_at_r
from autofaiss.metrics.reconstruction import quantize_vec_without_modifying_index, reconstruction_error
from autofaiss.utils.cast import cast_memory_to_bytes
from autofaiss.utils.decorators import Timeit


def compute_fast_metrics(
    embedding_reader: Union[np.ndarray, EmbeddingReader],
    index: faiss.Index,
    omp_threads: Optional[int] = None,
    query_max: Optional[int] = 1000,
) -> Dict:
    """compute query speed, size and reconstruction of an index"""
    infos: Dict[str, Union[str, int, float]] = {}

    size_bytes = get_index_size(index)
    infos["size in bytes"] = size_bytes

    if isinstance(embedding_reader, EmbeddingReader):
        # pylint: disable=bare-except
        query_embeddings, _ = next(embedding_reader(start=0, end=query_max, batch_size=query_max))
    else:
        query_embeddings = embedding_reader[:query_max]

    if omp_threads:
        faiss.omp_set_num_threads(1)
    speeds_ms = search_speed_test(index, query_embeddings, ksearch=40, timout_s=10.0)
    if omp_threads:
        faiss.omp_set_num_threads(omp_threads)

    infos.update(speeds_ms)

    # quantize query embeddings if the index uses quantization
    quantized_embeddings = quantize_vec_without_modifying_index(index, query_embeddings)
    rec_error = reconstruction_error(query_embeddings, quantized_embeddings)
    infos["reconstruction error %"] = 100 * rec_error

    infos["nb vectors"] = index.ntotal

    infos["vectors dimension"] = index.d

    infos["compression ratio"] = 4.0 * index.ntotal * index.d / size_bytes

    return infos


def compute_medium_metrics(
    embedding_reader: Union[np.ndarray, EmbeddingReader],
    index: faiss.Index,
    memory_available: Union[str, float],
    ground_truth: Optional[np.ndarray] = None,
    eval_item_ids: Optional[np.ndarray] = None,
) -> Dict[str, float]:
    """Compute recall@R and intersection recall@R of an index"""

    nb_test_points = 500

    if isinstance(embedding_reader, EmbeddingReader):
        query_embeddings, _ = next(embedding_reader(start=0, end=nb_test_points, batch_size=nb_test_points))
    else:
        embedding_block = embedding_reader

        query_embeddings = embedding_block[:nb_test_points]

    if ground_truth is None:
        if isinstance(embedding_reader, EmbeddingReader):
            ground_truth_path = f"{embedding_reader.embeddings_folder}/small_ground_truth_test.gt"
            fs, path = fsspec.core.url_to_fs(ground_truth_path, use_listings_cache=False)
            if not fs.exists(path):
                with Timeit("-> Compute small ground truth", indent=1):
                    ground_truth = get_ground_truth(
                        index.metric_type, embedding_reader, query_embeddings, memory_available
                    )

                    with fs.open(path, "wb") as gt_file:
                        np.save(gt_file, ground_truth)

            else:
                with Timeit("-> Load small ground truth", indent=1):
                    with fs.open(path, "rb") as gt_file:
                        ground_truth = np.load(gt_file)
        else:
            ground_truth = get_ground_truth(index.metric_type, embedding_block, query_embeddings, memory_available)

    with Timeit("-> Compute recalls", indent=1):
        one_recall = one_recall_at_r(query_embeddings, ground_truth, index, 40, eval_item_ids)
        intersection_recall = r_recall_at_r(query_embeddings, ground_truth, index, 40, eval_item_ids)

    infos: Dict[str, float] = {}

    infos["1-recall@20"] = one_recall[20 - 1]
    infos["1-recall@40"] = one_recall[40 - 1]
    infos["20-recall@20"] = intersection_recall[20 - 1]
    infos["40-recall@40"] = intersection_recall[40 - 1]

    return infos


def get_ground_truth(
    faiss_metric_type: int,
    embedding_reader: Union[np.ndarray, EmbeddingReader],
    query_embeddings: np.ndarray,
    memory_available: Union[str, float],
):
    """compute the ground truth (result with a perfect index) of the query on the embeddings"""

    dim = query_embeddings.shape[-1]

    if isinstance(embedding_reader, EmbeddingReader):
        perfect_index = MemEfficientFlatIndex(dim, faiss_metric_type)
        perfect_index.add_files(embedding_reader)
    else:
        perfect_index = faiss.IndexFlat(dim, faiss_metric_type)
        perfect_index.add(embedding_reader.astype("float32"))  # pylint: disable= no-value-for-parameter

    memory_available = cast_memory_to_bytes(memory_available) if isinstance(memory_available, str) else memory_available

    batch_size = int(min(memory_available, 10**9) / (dim * 4))  # at most 1GB of memory

    if isinstance(embedding_reader, EmbeddingReader):
        _, ground_truth = perfect_index.search_files(query_embeddings, k=40, batch_size=batch_size)
    else:
        _, ground_truth = perfect_index.search(query_embeddings, k=40)

    return ground_truth



""" functions to compare different indices """

import time

import numpy as np
from matplotlib import pyplot as plt
from tqdm import tqdm as tq

from autofaiss.indices.index_utils import format_speed_ms_per_query, get_index_size, speed_test_ms_per_query
from autofaiss.metrics.recalls import r_recall_at_r_single, one_recall_at_r_single
from autofaiss.utils.cast import cast_bytes_to_memory_string


def avg_speed_dict_ms_per_query(indices_dict, vectors, k_closest: int = 40, timeout_s: float = 5):
    """compute the average query speed of a dictionary of indices"""
    speed_dict = {}

    for index_key in indices_dict:
        speed = speed_test_ms_per_query(indices_dict[index_key], vectors, k_closest, timeout_s)
        speed_dict[index_key] = speed

    return speed_dict


def index_sizes_in_bytes_dict(indices_dict):
    """compute sizes of indices in a dictionary of indices"""
    size_dict = {}

    for index_key in indices_dict:
        size_dict[index_key] = get_index_size(indices_dict[index_key])

    return size_dict


def benchmark_index(
    indices_dict, gt_test, test_points, vectors_size_in_bytes, save_path=None, speed_dict=None, size_dict=None
):
    """
    Compute recall curves for the indices.
    """

    perfect_index_label = "perfect index"

    if perfect_index_label not in indices_dict:
        indices_dict[perfect_index_label] = None
        if speed_dict:
            speed_dict[perfect_index_label] = vectors_size_in_bytes

    k_max = gt_test.shape[1]

    plt.figure(figsize=(16, 8))

    k_values = np.arange(0, k_max + 1)

    avg_one_recall_at_r = {}
    avg_r_recall_at_r = {}

    timout_s = 5.0

    comp_size = vectors_size_in_bytes

    for index_key in tq(list(sorted(indices_dict.keys()))):
        if index_key not in indices_dict:
            continue

        index = indices_dict[index_key]

        if index_key == "Flat" or (index is None):
            y_r_recall_at_r = np.arange(1, k_max + 1)
            y_one_recall_at_r = np.ones(k_max)
            tot = 1
        else:
            y_r_recall_at_r = np.zeros(k_max)
            y_one_recall_at_r = np.zeros(k_max)
            tot = 0
            start_time = time.time()
            for i, item in enumerate(test_points):
                y_r_recall_at_r += np.array(r_recall_at_r_single(item, gt_test[i], index, k_max))
                y_one_recall_at_r += np.array(one_recall_at_r_single(item, gt_test[i], index, k_max))
                tot += 1
                if time.time() - start_time > timout_s and tot > 150:
                    break

        avg_r_recall_at_r[index_key] = y_r_recall_at_r / tot
        avg_one_recall_at_r[index_key] = y_one_recall_at_r / tot

    info_string = {index_key: "" for index_key in indices_dict}

    initial_size_string = cast_bytes_to_memory_string(comp_size)

    for index_key in indices_dict:
        if index_key in speed_dict:
            info_string[index_key] += f"avg speed: {format_speed_ms_per_query(speed_dict[index_key])}, "
        if index_key in size_dict:
            info_string[index_key] += (
                f"(Size: {cast_bytes_to_memory_string(size_dict[index_key])} "
                f"({(100*size_dict[index_key]/comp_size):.1f}% of {initial_size_string})"
            )

    plt.subplot(121)

    for index_key in sorted(indices_dict.keys()):
        if index_key not in indices_dict:
            continue

        label = f"{index_key:<30} Index, {info_string[index_key]}"

        plt.plot(k_values, np.concatenate(([0], avg_r_recall_at_r[index_key])), label=label)

    plt.xlabel("k, number of nearests items")
    plt.ylabel("k-recall@k")
    plt.vlines(40, 0, k_max)
    plt.legend()
    plt.tight_layout()

    plt.subplot(122)

    for index_key in sorted(indices_dict.keys()):
        if index_key not in indices_dict:
            continue

        label = f"{index_key:<30} Index, {info_string[index_key]}"

        plt.plot(k_values, np.concatenate(([0], 100 * avg_one_recall_at_r[index_key])), label=label)

    plt.xlabel("k, number of nearests items")
    plt.ylabel("1-Recall@k")
    plt.vlines(100, 0, k_max)
    plt.legend()
    plt.tight_layout()

    if save_path:
        plt.savefig(save_path)

    plt.show()



from autofaiss.indices.distributed import _batch_loader


def test_batch_loader():
    for input_size in range(2, 500):
        for output_size in range(1, input_size):
            batches = list(_batch_loader(nb_batches=output_size, total_size=input_size))
            # test output size is expected
            assert len(batches) == output_size
            # test no empty batch
            assert all(batch[1] <= input_size - 1 for batch in batches)
            # test on continuous between batches
            assert all(prev_end == next_start for (_, _, prev_end), (_, next_start, _) in zip(batches, batches[1:]))
            # test last element is covered
            assert batches[-1][2] >= input_size
            # test range sum
            assert sum(end - start for _, start, end in batches) == input_size



""" Common functions to build an index """

import logging
from typing import Dict, Optional, Tuple, Union, Callable, Any
import uuid
import re
import os
import tempfile

import fsspec
import faiss
import pandas as pd
from embedding_reader import EmbeddingReader

from autofaiss.external.optimize import optimize_and_measure_index, get_optimal_batch_size
from autofaiss.indices.index_utils import set_search_hyperparameters, initialize_direct_map, load_index
from autofaiss.utils.path import make_path_absolute
from autofaiss.utils.cast import cast_bytes_to_memory_string


logger = logging.getLogger("autofaiss")


def get_write_ids_df_to_parquet_fn(ids_root_dir: str) -> Callable[[pd.DataFrame, int], None]:
    """Create function to write ids from Pandas dataframe to parquet"""

    def _write_ids_df_to_parquet_fn(ids: pd.DataFrame, batch_id: int):
        filename = f"part-{batch_id:08d}-{uuid.uuid1()}.parquet"
        output_file = os.path.join(ids_root_dir, filename)  # type: ignore
        with fsspec.open(output_file, "wb") as f:
            logger.debug(f"Writing id DataFrame to file {output_file}")
            ids.to_parquet(f, index=False)

    return _write_ids_df_to_parquet_fn


def get_optimize_index_fn(
    embedding_reader: EmbeddingReader,
    index_key: str,
    index_path: Optional[str],
    index_infos_path: Optional[str],
    use_gpu: bool,
    save_on_disk: bool,
    max_index_query_time_ms: float,
    min_nearest_neighbors_to_retrieve: int,
    make_direct_map: bool,
    index_param: Optional[str],
) -> Callable[[faiss.Index, str], Dict]:
    """Create function to optimize index by choosing best hyperparameters and calculating metrics"""

    def _optimize_index_fn(index: faiss.Index, index_suffix: str):
        if make_direct_map:
            initialize_direct_map(index)

        cur_index_path = make_path_absolute(index_path) + index_suffix if index_path else None
        cur_index_infos_path = make_path_absolute(index_infos_path) + index_suffix if index_infos_path else None
        if any(re.findall(r"OPQ\d+_\d+,IVF\d+_HNSW\d+,PQ\d+", index_key)):
            set_search_hyperparameters(index, f"nprobe={64},efSearch={128},ht={2048}", use_gpu)
        metric_infos = optimize_and_measure_index(
            embedding_reader,
            index,
            cur_index_infos_path,
            index_key,
            index_param,
            cur_index_path,
            max_index_query_time_ms=max_index_query_time_ms,
            min_nearest_neighbors_to_retrieve=min_nearest_neighbors_to_retrieve,
            save_on_disk=save_on_disk,
            use_gpu=use_gpu,
        )
        return metric_infos

    return _optimize_index_fn


def add_embeddings_to_index_local(
    embedding_reader: EmbeddingReader,
    trained_index_or_path: Union[faiss.Index, str],
    memory_available_for_adding: str,
    embedding_ids_df_handler: Optional[Callable[[pd.DataFrame, int], Any]] = None,
    index_optimizer: Optional[Callable] = None,
    add_embeddings_with_ids: bool = False,
) -> Tuple[Optional[faiss.Index], Optional[Dict[str, str]]]:
    """Add embeddings to index from driver"""

    vec_dim = embedding_reader.dimension
    batch_size = get_optimal_batch_size(vec_dim, memory_available_for_adding)
    logger.info(
        f"Using a batch size of {batch_size} (memory overhead {cast_bytes_to_memory_string(batch_size * vec_dim * 4)})"
    )

    with tempfile.TemporaryDirectory() as tmp_dir:
        if isinstance(trained_index_or_path, str):
            local_index_path = os.path.join(tmp_dir, "index")
            trained_index = load_index(trained_index_or_path, local_index_path)
        else:
            trained_index = trained_index_or_path
        for batch_id, (vec_batch, ids_batch) in enumerate(embedding_reader(batch_size=batch_size)):
            if add_embeddings_with_ids:
                trained_index.add_with_ids(vec_batch, ids_batch["i"].to_numpy())
            else:
                trained_index.add(vec_batch)
            if embedding_ids_df_handler:
                embedding_ids_df_handler(ids_batch, batch_id)
        metric_infos = index_optimizer(trained_index, "") if index_optimizer else None  # type: ignore
        return trained_index, metric_infos



""" function to compute the reconstruction error """

from typing import Optional

import numpy as np
import faiss


def reconstruction_error(before, after, avg_norm_before: Optional[float] = None) -> float:
    """Computes the average reconstruction error"""
    diff = np.mean(np.linalg.norm(after - before, axis=1))
    if avg_norm_before is None:
        avg_norm_before = np.mean(np.linalg.norm(before, axis=1))
    return diff / avg_norm_before


def quantize_vec_without_modifying_index(index: faiss.Index, vecs: np.ndarray) -> np.ndarray:
    """Quantizes a batch of vectors if the index given uses quantization"""

    try:
        return index.sa_decode(index.sa_encode(vecs))
    except (TypeError, RuntimeError):  # error if the index doesn't use quantization
        return vecs



""" function to cast variables in others """


import re
from math import floor
from typing import Union

import faiss


def cast_memory_to_bytes(memory_string: str) -> float:
    """
    Parse a memory string and returns the number of bytes
    >>> cast_memory_to_bytes("16B")
    16
    >>> cast_memory_to_bytes("16G") == 16*1024*1024*1024
    True
    """

    conversion = {unit: (2**10) ** i for i, unit in enumerate("BKMGTPEZ")}

    number_match = r"([0-9]*\.[0-9]+|[0-9]+)"
    unit_match = "("
    for unit in conversion:
        if unit != "B":
            unit_match += unit + "B|"
    for unit in conversion:
        unit_match += unit + "|"
    unit_match = unit_match[:-1] + ")"

    matching_groups = re.findall(number_match + unit_match, memory_string, re.IGNORECASE)

    if matching_groups and len(matching_groups) == 1 and "".join(matching_groups[0]) == memory_string:
        group = matching_groups[0]
        return float(group[0]) * conversion[group[1][0].upper()]

    raise ValueError(f"Unknown format for memory string: {memory_string}")


def cast_bytes_to_memory_string(num_bytes: float) -> str:
    """
    Cast a number of bytes to a readable string
    """

    suffix = "B"
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num_bytes) < 1024.0:
            return "%3.1f%s%s" % (num_bytes, unit, suffix)  # pylint: disable=consider-using-f-string
        num_bytes /= 1024.0
    return "%.1f%s%s" % (num_bytes, "Y", suffix)  # pylint: disable=consider-using-f-string


def to_faiss_metric_type(metric_type: Union[str, int]) -> int:
    """convert metric_type string/enum to faiss enum of the distance metric"""

    if metric_type in ["ip", "IP", faiss.METRIC_INNER_PRODUCT]:
        return faiss.METRIC_INNER_PRODUCT
    elif metric_type in ["l2", "L2", faiss.METRIC_L2]:
        return faiss.METRIC_L2
    else:
        raise ValueError("Metric currently not supported")


def to_readable_time(seconds: float, rounding: bool = False) -> str:
    """cast time in seconds to readable string"""

    initial_seconds = seconds

    hours = int(floor(seconds // 3600))
    seconds -= 3600 * hours
    minutes = int(floor(seconds // 60))
    seconds -= 60 * minutes

    if rounding:
        if hours:
            return f"{initial_seconds/3600:3.1f} hours"
        if minutes:
            return f"{initial_seconds/60:3.1f} minutes"
        return f"{initial_seconds:3.1f} seconds"

    time_str = ""
    if hours:
        time_str += f"{hours:d} hours "
    if hours or minutes:
        time_str += f"{minutes:d} minutes "
    time_str += f"{seconds:.2f} seconds"

    return time_str



"""
Index metadata for Faiss indices.
"""

import re
from enum import Enum
from math import ceil

from autofaiss.utils.cast import cast_bytes_to_memory_string
from autofaiss.external.descriptions import (
    INDEX_DESCRIPTION_BLOCKS,
    IndexBlock,
    TUNABLE_PARAMETERS_DESCRIPTION_BLOCKS,
    TunableParam,
)


class IndexType(Enum):
    FLAT = 0
    HNSW = 1
    OPQ_IVF_PQ = 2
    OPQ_IVF_HNSW_PQ = 3
    PAD_IVF_HNSW_PQ = 4
    NOT_SUPPORTED = 5
    IVF_FLAT = 6


class IndexMetadata:
    """
    Class to compute index metadata given the index_key, the number of vectors and their dimension.

    Note: We don't create classes for each index type in order to keep the code simple.
    """

    def __init__(self, index_key: str, nb_vectors: int, dim_vector: int, make_direct_map: bool = False):
        self.index_key = index_key
        self.nb_vectors = nb_vectors
        self.dim_vector = dim_vector
        self.fast_description = ""
        self.description_blocs = []
        self.tunable_params = []
        self.params = {}
        self.make_direct_map = make_direct_map

        params = [int(x) for x in re.findall(r"\d+", index_key)]

        if any(re.findall(r"OPQ\d+_\d+,IVF\d+,PQ\d+", index_key)):
            self.index_type = IndexType.OPQ_IVF_PQ
            self.fast_description = "An inverted file index (IVF) with quantization and OPQ preprocessing."
            self.description_blocs = [IndexBlock.IVF, IndexBlock.PQ, IndexBlock.OPQ]
            self.tunable_params = [TunableParam.NPROBE, TunableParam.HT]

            self.params["pq"] = params[3]
            self.params["nbits"] = params[4] if len(params) == 5 else 8  # default value
            self.params["ncentroids"] = params[2]
            self.params["out_d"] = params[1]
            self.params["M_OPQ"] = params[0]

        elif any(re.findall(r"OPQ\d+_\d+,IVF\d+_HNSW\d+,PQ\d+", index_key)):
            self.index_type = IndexType.OPQ_IVF_HNSW_PQ
            self.fast_description = "An inverted file index (IVF) with quantization, OPQ preprocessing, and HNSW index."
            self.description_blocs = [IndexBlock.IVF_HNSW, IndexBlock.HNSW, IndexBlock.PQ, IndexBlock.OPQ]
            self.tunable_params = [TunableParam.NPROBE, TunableParam.EFSEARCH, TunableParam.HT]

            self.params["M_HNSW"] = params[3]
            self.params["pq"] = params[4]
            self.params["nbits"] = params[5] if len(params) == 6 else 8  # default value
            self.params["ncentroids"] = params[2]
            self.params["out_d"] = params[1]
            self.params["M_OPQ"] = params[0]

        elif any(re.findall(r"Pad\d+,IVF\d+_HNSW\d+,PQ\d+", index_key)):
            self.index_type = IndexType.PAD_IVF_HNSW_PQ
            self.fast_description = (
                "An inverted file index (IVF) with quantization, a padding on input vectors, and HNSW index."
            )
            self.description_blocs = [IndexBlock.IVF_HNSW, IndexBlock.HNSW, IndexBlock.PQ, IndexBlock.PAD]
            self.tunable_params = [TunableParam.NPROBE, TunableParam.EFSEARCH, TunableParam.HT]

            self.params["out_d"] = params[0]
            self.params["M_HNSW"] = params[2]
            self.params["pq"] = params[3]
            self.params["nbits"] = params[4] if len(params) == 5 else 8  # default value
            self.params["ncentroids"] = params[1]

        elif any(re.findall(r"HNSW\d+", index_key)):
            self.index_type = IndexType.HNSW
            self.fast_description = "An HNSW index."
            self.description_blocs = [IndexBlock.HNSW]
            self.tunable_params = [TunableParam.EFSEARCH]

            self.params["M_HNSW"] = params[0]

        elif any(re.findall(r"IVF\d+,Flat", index_key)):
            self.index_type = IndexType.IVF_FLAT
            self.fast_description = "An inverted file index (IVF) with no quantization"
            self.description_blocs = [IndexBlock.IVF]
            self.tunable_params = [TunableParam.NPROBE]

            self.params["ncentroids"] = params[0]

        elif index_key == "Flat":
            self.index_type = IndexType.FLAT
            self.fast_description = "A simple flat index."
            self.description_blocs = [IndexBlock.FLAT]
            self.tunable_params = []

        else:
            self.index_type = IndexType.NOT_SUPPORTED
            self.fast_description = "No description for this index, feel free to contribute :)"
            self.description_blocs = []
            self.tunable_params = []

    def get_index_type(self) -> IndexType:
        """
        return the index type.
        """
        return self.index_type

    def estimated_index_size_in_bytes(self) -> int:
        """
        Compute the estimated size of the index in bytes.
        """

        if self.index_type == IndexType.FLAT:
            return self.nb_vectors * self.dim_vector * 4

        if self.index_type == IndexType.HNSW:
            # M bidirectional links per vector in the HNSW graph
            hnsw_graph_in_bytes = self.nb_vectors * self.params["M_HNSW"] * 2 * 4
            vectors_size_in_bytes = self.nb_vectors * self.dim_vector * 4
            return vectors_size_in_bytes + hnsw_graph_in_bytes

        if self.index_type in [IndexType.OPQ_IVF_PQ, IndexType.OPQ_IVF_HNSW_PQ, IndexType.PAD_IVF_HNSW_PQ]:
            direct_map_overhead = 8 * self.nb_vectors if self.make_direct_map else 0

            # We neglict the size of the OPQ table for the moment.
            code_size = ceil(self.params["pq"] * self.params["nbits"] / 8)

            # https://github.com/facebookresearch/faiss/blob/main/faiss/invlists/InvertedLists.h#L193
            embedding_id_byte = 8

            vector_size_byte = code_size + embedding_id_byte

            vectors_size_in_bytes = self.nb_vectors * vector_size_byte
            centroid_size_in_bytes = self.params["ncentroids"] * self.dim_vector * 4

            total_size_in_byte = direct_map_overhead + vectors_size_in_bytes + centroid_size_in_bytes

            if self.index_type in [IndexType.OPQ_IVF_HNSW_PQ, IndexType.PAD_IVF_HNSW_PQ]:
                total_size_in_byte += self.params["ncentroids"] * self.params["M_HNSW"] * 2 * 4

            if self.index_type in [IndexType.OPQ_IVF_PQ, IndexType.OPQ_IVF_HNSW_PQ]:
                total_size_in_byte += self.params["M_OPQ"] * self.params["out_d"] * 4

            return total_size_in_byte

        if self.index_type == IndexType.IVF_FLAT:
            direct_map_overhead = 8 * self.nb_vectors if self.make_direct_map else 0
            vectors_size_in_bytes = self.nb_vectors * self.dim_vector * 4
            centroid_size_in_bytes = self.params["ncentroids"] * self.dim_vector * 4

            return direct_map_overhead + vectors_size_in_bytes + centroid_size_in_bytes

        return -1

    def get_index_description(self, tunable_parameters_infos=False) -> str:
        """
        Gives a generic description of the index.
        """

        description = self.fast_description

        if self.index_type == IndexType.NOT_SUPPORTED:
            return description

        description += "\n"
        index_size_string = cast_bytes_to_memory_string(self.estimated_index_size_in_bytes())
        description += f"The size of the index should be around {index_size_string}.\n\n"
        description += "\n".join(INDEX_DESCRIPTION_BLOCKS[desc] for desc in self.description_blocs) + "\n\n"

        if tunable_parameters_infos:
            if not self.tunable_params:
                description += "No parameters can be tuned to find a query speed VS recall tradeoff\n\n"
            else:
                description += "List of parameters that can be tuned to find a query speed VS recall tradeoff:\n"
                description += (
                    "\n".join(TUNABLE_PARAMETERS_DESCRIPTION_BLOCKS[desc] for desc in self.tunable_params) + "\n\n"
                )

        description += """
For all indices except the flat index, the query speed can be adjusted.
The lower the speed limit the lower the recall. With a looser constraint
on the query time, the recall can be higher, but it is limited by the index
structure (if there is quantization for instance).
"""
        return description

    def compute_memory_necessary_for_training(self, nb_training_vectors: int) -> float:
        """
        Function that computes the memory necessary to train an index with nb_training_vectors vectors
        """
        if self.index_type == IndexType.FLAT:
            return 0
        elif self.index_type == IndexType.IVF_FLAT:
            return self.compute_memory_necessary_for_ivf_flat(nb_training_vectors)
        elif self.index_type == IndexType.HNSW:
            return 0
        elif self.index_type == IndexType.OPQ_IVF_PQ:
            return self.compute_memory_necessary_for_opq_ivf_pq(nb_training_vectors)
        elif self.index_type == IndexType.OPQ_IVF_HNSW_PQ:
            return self.compute_memory_necessary_for_opq_ivf_hnsw_pq(nb_training_vectors)
        elif self.index_type == IndexType.PAD_IVF_HNSW_PQ:
            return self.compute_memory_necessary_for_pad_ivf_hnsw_pq(nb_training_vectors)
        else:
            return 500 * 10**6

    def compute_memory_necessary_for_ivf_flat(self, nb_training_vectors: int):
        """Compute the memory estimation for index type IVF_FLAT."""
        ivf_memory_in_bytes = self._get_ivf_training_memory_usage_in_bytes()
        # TODO: remove x1.5 when estimation is correct
        return self._get_vectors_training_memory_usage_in_bytes(nb_training_vectors) * 1.5 + ivf_memory_in_bytes

    def compute_memory_necessary_for_opq_ivf_pq(self, nb_training_vectors: int) -> float:
        """Compute the memory estimation for index type OPQ_IVF_PQ."""
        # TODO: remove x1.5 when estimation is correct
        return (
            self._get_vectors_training_memory_usage_in_bytes(nb_training_vectors) * 1.5
            + self._get_opq_training_memory_usage_in_bytes(nb_training_vectors)
            + self._get_ivf_training_memory_usage_in_bytes()
            + self._get_pq_training_memory_usage_in_bytes()
        )

    def compute_memory_necessary_for_opq_ivf_hnsw_pq(self, nb_training_vectors: int) -> float:
        """Compute the memory estimation for index type OPQ_IVF_HNSW_PQ."""
        # TODO: remove x1.5 when estimation is correct
        return (
            self._get_vectors_training_memory_usage_in_bytes(nb_training_vectors) * 1.5
            + self._get_opq_training_memory_usage_in_bytes(nb_training_vectors)
            + self._get_ivf_training_memory_usage_in_bytes()
            + self._get_ivf_hnsw_training_memory_usage_in_bytes()
            + self._get_pq_training_memory_usage_in_bytes()
        )

    def compute_memory_necessary_for_pad_ivf_hnsw_pq(self, nb_training_vectors: int):
        """Compute the memory estimation for index type PAD_IVF_HNSW_PQ."""
        # TODO: remove x1.5 when estimation is correct
        return (
            self._get_vectors_training_memory_usage_in_bytes(nb_training_vectors) * 1.5
            + self._get_ivf_training_memory_usage_in_bytes()
            + self._get_ivf_hnsw_training_memory_usage_in_bytes()
            + self._get_pq_training_memory_usage_in_bytes()
        )

    def _get_vectors_training_memory_usage_in_bytes(self, nb_training_vectors: int):
        """Get vectors memory estimation in bytes."""
        return 4.0 * self.dim_vector * nb_training_vectors

    def _get_ivf_training_memory_usage_in_bytes(self):
        """Get IVF memory estimation in bytes."""
        return 4.0 * self.params["ncentroids"] * self.dim_vector

    def _get_ivf_hnsw_training_memory_usage_in_bytes(self):
        """Get HNSW followed by IVF memory estimation in bytes."""

        hnsw_graph_in_bytes = self.params["ncentroids"] * self.params["M_HNSW"] * 2 * 4
        vectors_size_in_bytes = self.params["ncentroids"] * self.dim_vector * 4
        return vectors_size_in_bytes + hnsw_graph_in_bytes

    def _get_opq_training_memory_usage_in_bytes(self, nb_training_vectors: int):
        """Get OPQ memory estimation in bytes."""
        # see OPQMatrix::train on https://github.com/facebookresearch/faiss/blob/main/faiss/VectorTransform.cpp#L987
        d_in, d_out, code_size = self.dim_vector, self.params["out_d"], self.params["M_OPQ"]
        n = min(256 * 256, nb_training_vectors)
        d = max(d_in, d_out)
        d2 = d_out
        xproj = d2 * n
        pq_recons = d2 * n
        xxr = d * n
        tmp = d * d * 4
        mem_code_size = code_size * n
        opq_memory_in_bytes = (xproj + pq_recons + xxr + tmp) * 4.0 + mem_code_size * 1.0
        return opq_memory_in_bytes

    def _get_pq_training_memory_usage_in_bytes(self):
        """Get PQ memory estimation in bytes."""
        return ceil(self.params["pq"] * self.params["nbits"] / 8)


def compute_memory_necessary_for_training_wrapper(nb_training_vectors: int, index_key: str, dim_vector: int):
    # nb_vectors is useless for training memory estimation, so just put -1
    return IndexMetadata(index_key, -1, dim_vector).compute_memory_necessary_for_training(
        nb_training_vectors=nb_training_vectors
    )



# pylint: disable=unused-import,missing-docstring

from autofaiss.external.quantize import build_index, score_index, tune_index, build_partitioned_indexes

from autofaiss.version import __author__, __version__
